import json
import pathlib
from typing import Any, Dict, List, TYPE_CHECKING

from sinter._task_stats import TaskStats
from sinter._task import Task
from sinter._decoding import AnonTaskStats

if TYPE_CHECKING:
    import sinter


class ExistingData:
    def __init__(self):
        self.data: Dict[str, TaskStats] = {}

    def stats_for(self, case: Task) -> AnonTaskStats:
        if isinstance(case, Task):
            key = case.strong_id()
        else:
            raise NotImplementedError(f'{type(case)}')
        if key not in self.data:
            return AnonTaskStats()
        return self.data[key].to_anon_stats()

    def add_sample(self, sample: TaskStats) -> None:
        k = sample.strong_id
        if k in self.data:
            self.data[k] += sample
        else:
            self.data[k] = sample

    def __iadd__(self, other: 'ExistingData') -> 'ExistingData':
        for sample in other.data.values():
            self.add_sample(sample)
        return self

    @staticmethod
    def from_file(path_or_file: Any) -> 'ExistingData':
        expected_fields = {
            "shots",
            "discards",
            "errors",
            "seconds",
            "strong_id",
            "decoder",
            "json_metadata",
        }
        # Import is done locally to reduce cost of importing sinter.
        import csv
        if isinstance(path_or_file, (str, pathlib.Path)):
            with open(path_or_file) as csvfile:
                return ExistingData.from_file(csvfile)
        reader = csv.DictReader(path_or_file)
        reader.fieldnames = [e.strip() for e in reader.fieldnames]
        actual_fields = set(reader.fieldnames)
        if actual_fields != expected_fields:
            raise ValueError(
                f"Bad CSV data. "
                f"Got columns {sorted(actual_fields)!r} "
                f"but expected columns {sorted(expected_fields)!r}")
        result = ExistingData()
        for row in reader:
            result.add_sample(TaskStats(
                shots=int(row['shots']),
                discards=int(row['discards']),
                errors=int(row['errors']),
                seconds=float(row['seconds']),
                strong_id=row['strong_id'],
                decoder=row['decoder'],
                json_metadata=json.loads(row['json_metadata']),
            ))
        return result


def stats_from_csv_files(*paths_or_files: Any) -> List['sinter.TaskStats']:
    result = ExistingData()
    for p in paths_or_files:
        result += ExistingData.from_file(p)
    return list(result.data.values())



import contextlib
import dataclasses
import pathlib
from typing import Any
from typing import Callable, Iterator, Optional, Union, Iterable, List, TYPE_CHECKING, Tuple, Dict

import math
import numpy as np
import stim

from sinter._collection_options import CollectionOptions
from sinter._csv_out import CSV_HEADER
from sinter._collection_work_manager import CollectionWorkManager
from sinter._existing_data import ExistingData
from sinter._printer import ThrottledProgressPrinter
from sinter._task_stats import TaskStats

if TYPE_CHECKING:
    import sinter


@dataclasses.dataclass(frozen=True)
class Progress:
    new_stats: Tuple[TaskStats, ...]
    status_message: str


def iter_collect(*,
                 num_workers: int,
                 tasks: Union[Iterator['sinter.Task'],
                              Iterable['sinter.Task']],
                 hint_num_tasks: Optional[int] = None,
                 additional_existing_data: Optional[ExistingData] = None,
                 max_shots: Optional[int] = None,
                 max_errors: Optional[int] = None,
                 decoders: Optional[Iterable[str]] = None,
                 max_batch_seconds: Optional[int] = None,
                 max_batch_size: Optional[int] = None,
                 start_batch_size: Optional[int] = None,
                 custom_decoders: Optional[Dict[str, 'sinter.Decoder']] = None,
                 ) -> Iterator['sinter.Progress']:
    """Collects error correction statistics using multiple worker processes.

    Note: if max_batch_size and max_batch_seconds are both not used (or
    explicitly set to None), a default batch-size-limiting mechanism will be
    chosen.

    Args:
        num_workers: The number of worker processes to use.
        tasks: Decoding problems to sample.
        hint_num_tasks: If `tasks` is an iterator or a generator, its length
            can be given here so that progress printouts can say how many cases
            are left.
        additional_existing_data: Defaults to None (no additional data).
            Statistical data that has already been collected, in addition to
            anything included in each task's `previous_stats` field.
        decoders: Defaults to None (specified by each Task). The names of the
            decoders to use on each Task. It must either be the case that each
            Task specifies a decoder and this is set to None, or this is an
            iterable and each Task has its decoder set to None.
        max_shots: Defaults to None (unused). Stops the sampling process
            after this many samples have been taken from the circuit.
        max_errors: Defaults to None (unused). Stops the sampling process
            after this many errors have been seen in samples taken from the
            circuit. The actual number sampled errors may be larger due to
            batching.
        start_batch_size: Defaults to None (collector's choice). The very
            first shots taken from the circuit will use a batch of this
            size, and no other batches will be taken in parallel. Once this
            initial fact finding batch is done, batches can be taken in
            parallel and the normal batch size limiting processes take over.
        max_batch_size: Defaults to None (unused). Limits batches from
            taking more than this many shots at once. For example, this can
            be used to ensure memory usage stays below some limit.
        max_batch_seconds: Defaults to None (unused). When set, the recorded
            data from previous shots is used to estimate how much time is
            taken per shot. This information is then used to predict the
            biggest batch size that can finish in under the given number of
            seconds. Limits each batch to be no larger than that.
        custom_decoders: Custom decoders that can be used if requested by name.
            If not specified, only decoders built into sinter, such as
            'pymatching' and 'fusion_blossom', can be used.

    Yields:
        sinter.SamplerStats values recording incremental statistical data
        as it is collected by workers.
    """
    if hint_num_tasks is None:
        try:
            # noinspection PyTypeChecker
            hint_num_tasks = len(tasks)
        except TypeError:
            pass

    with CollectionWorkManager(
            tasks_iter=iter(tasks),
            global_collection_options=CollectionOptions(
                max_shots=max_shots,
                max_errors=max_errors,
                max_batch_seconds=max_batch_seconds,
                start_batch_size=start_batch_size,
                max_batch_size=max_batch_size,
            ),
            decoders=decoders,
            additional_existing_data=additional_existing_data,
            custom_decoders=custom_decoders) as manager:
        try:
            yield Progress(
                new_stats=(),
                status_message="Starting workers..."
            )
            manager.start_workers(num_workers)

            yield Progress(
                new_stats=(),
                status_message="Finding work..."
            )
            manager.fill_work_queue()
            yield Progress(
                new_stats=(),
                status_message=manager.status(num_circuits=hint_num_tasks)
            )

            while manager.fill_work_queue():
                # Wait for a worker to finish a job.
                sample = manager.wait_for_next_sample()
                manager.fill_work_queue()

                # Report the incremental results.
                yield Progress(
                    new_stats=(sample,) if sample.shots > 0 else (),
                    status_message=manager.status(num_circuits=hint_num_tasks),
                )
        except KeyboardInterrupt:
            yield Progress(
                new_stats=(),
                status_message='KeyboardInterrupt',
            )
            raise


def collect(*,
            num_workers: int,
            tasks: Union[Iterator['sinter.Task'], Iterable['sinter.Task']],
            existing_data_filepaths: Iterable[Union[str, pathlib.Path]] = (),
            save_resume_filepath: Union[None, str, pathlib.Path] = None,
            progress_callback: Optional[Callable[['sinter.Progress'], None]] = None,
            max_shots: Optional[int] = None,
            max_errors: Optional[int] = None,
            decoders: Optional[Iterable[str]] = None,
            max_batch_seconds: Optional[int] = None,
            max_batch_size: Optional[int] = None,
            start_batch_size: Optional[int] = None,
            print_progress: bool = False,
            hint_num_tasks: Optional[int] = None,
            custom_decoders: Optional[Dict[str, 'sinter.Decoder']] = None,
            ) -> List['sinter.TaskStats']:
    """
    Args:
        num_workers: The number of worker processes to use.
        tasks: Decoding problems to sample.
        save_resume_filepath: Defaults to None (unused). If set to a filepath,
            results will be saved to that file while they are collected. If the
            python interpreter is stopped or killed, calling this method again
            with the same save_resume_filepath will load the previous results
            from the file so it can resume where it left off.

            The stats in this file will be counted in addition to each task's
            previous_stats field (as opposed to overriding the field).
        existing_data_filepaths: CSV data saved to these files will be loaded,
            included in the returned results, and count towards things like
            max_shots and max_errors.
        progress_callback: Defaults to None (unused). If specified, then each
            time new sample statistics are acquired from a worker this method
            will be invoked with the new sinter.SamplerStats.
        hint_num_tasks: If `tasks` is an iterator or a generator, its length
            can be given here so that progress printouts can say how many cases
            are left.
        decoders: Defaults to None (specified by each Task). The names of the
            decoders to use on each Task. It must either be the case that each
            Task specifies a decoder and this is set to None, or this is an
            iterable and each Task has its decoder set to None.
        max_shots: Defaults to None (unused). Stops the sampling process
            after this many samples have been taken from the circuit.
        max_errors: Defaults to None (unused). Stops the sampling process
            after this many errors have been seen in samples taken from the
            circuit. The actual number sampled errors may be larger due to
            batching.
        start_batch_size: Defaults to None (collector's choice). The very
            first shots taken from the circuit will use a batch of this
            size, and no other batches will be taken in parallel. Once this
            initial fact finding batch is done, batches can be taken in
            parallel and the normal batch size limiting processes take over.
        max_batch_size: Defaults to None (unused). Limits batches from
            taking more than this many shots at once. For example, this can
            be used to ensure memory usage stays below some limit.
        print_progress: When True, progress is printed to stderr while
            collection runs.
        max_batch_seconds: Defaults to None (unused). When set, the recorded
            data from previous shots is used to estimate how much time is
            taken per shot. This information is then used to predict the
            biggest batch size that can finish in under the given number of
            seconds. Limits each batch to be no larger than that.
        custom_decoders: Custom decoders that can be used if requested by name.
            If not specified, only decoders built into sinter, such as
            'pymatching' and 'fusion_blossom', can be used.

    Returns:
        A list of sample statistics, one from each problem. The list is not in
        any specific order. This is the same data that would have been written
        to a CSV file, but aggregated so that each problem has exactly one
        sample statistic instead of potentially multiple.
    """
    # Load existing data.
    additional_existing_data = ExistingData()
    for existing in existing_data_filepaths:
        additional_existing_data += ExistingData.from_file(existing)

    if save_resume_filepath in existing_data_filepaths:
        raise ValueError("save_resume_filepath in existing_data_filepaths")

    progress_printer = ThrottledProgressPrinter(
        outs=[],
        print_progress=print_progress,
        min_progress_delay=1,
    )
    with contextlib.ExitStack() as exit_stack:
        # Open save/resume file.
        if save_resume_filepath is not None:
            save_resume_filepath = pathlib.Path(save_resume_filepath)
            if save_resume_filepath.exists():
                additional_existing_data += ExistingData.from_file(save_resume_filepath)
                save_resume_file = exit_stack.enter_context(
                        open(save_resume_filepath, 'a'))  # type: ignore
            else:
                save_resume_filepath.parent.mkdir(exist_ok=True)
                save_resume_file = exit_stack.enter_context(
                        open(save_resume_filepath, 'w'))  # type: ignore
                print(CSV_HEADER, file=save_resume_file, flush=True)
        else:
            save_resume_file = None

        # Collect data.
        result = ExistingData()
        result.data = dict(additional_existing_data.data)
        for progress in iter_collect(
            num_workers=num_workers,
            max_shots=max_shots,
            max_errors=max_errors,
            max_batch_seconds=max_batch_seconds,
            start_batch_size=start_batch_size,
            max_batch_size=max_batch_size,
            decoders=decoders,
            tasks=tasks,
            hint_num_tasks=hint_num_tasks,
            additional_existing_data=additional_existing_data,
            custom_decoders=custom_decoders,
        ):
            for stats in progress.new_stats:
                result.add_sample(stats)
                if save_resume_file is not None:
                    print(stats.to_csv_line(), file=save_resume_file, flush=True)
            if print_progress:
                progress_printer.show_latest_progress(progress.status_message)
            if progress_callback is not None:
                progress_callback(progress)
        if print_progress:
            progress_printer.flush()
        return list(result.data.values())


def post_selection_mask_from_predicate(
    circuit_or_dem: Union[stim.Circuit, stim.DetectorErrorModel],
    *,
    postselected_detectors_predicate: Callable[[int, Any, Tuple[float, ...]], bool],
    metadata: Any,
) -> np.ndarray:
    num_dets = circuit_or_dem.num_detectors
    post_selection_mask = np.zeros(dtype=np.uint8, shape=math.ceil(num_dets / 8))
    for k, coord in circuit_or_dem.get_detector_coordinates().items():
        if postselected_detectors_predicate(k, metadata, coord):
            post_selection_mask[k // 8] |= 1 << (k % 8)
    return post_selection_mask


def post_selection_mask_from_4th_coord(dem: Union[stim.Circuit, stim.DetectorErrorModel]) -> np.ndarray:
    num_dets = dem.num_detectors
    post_selection_mask = np.zeros(dtype=np.uint8, shape=math.ceil(num_dets / 8))
    for k, coord in dem.get_detector_coordinates().items():
        if len(coord) >= 4 and coord[3]:
            post_selection_mask[k // 8] |= 1 << (k % 8)
    return post_selection_mask



import contextlib
import io
import pathlib
import tempfile

from sinter._main import main


def test_main_combine():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
       600,       101,       220,    3.00,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
         9,         5,         4,    6.00,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
"""

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                str(d / "input.csv"),
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
      1200,       202,       440,    6.00,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
        18,        10,         8,    12.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
"""


def test_order_flag():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,   strong_id,json_metadata
1000, 100,   4,       2.0,    pymatching,deadbeef0,"{""d"":19}"
2000, 300,   3,       3.0,    pymatching,deadbeef1,"{""d"":9}"
3000, 200,   2000,    5.0,    pymatching,deadbeef2,"{""d"":200}"
4000, 100,   1,       7.0,    pymatching,deadbeef3,"{""d"":3}"
5000, 100,   0,       11,     pymatching,deadbeef4,"{""d"":5}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                "--order",
                "preserve",
                str(d / "input.csv"),
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
      2000,       200,         8,    4.00,pymatching,deadbeef0,"{""d"":19}"
      4000,       600,         6,    6.00,pymatching,deadbeef1,"{""d"":9}"
      6000,       400,      4000,    10.0,pymatching,deadbeef2,"{""d"":200}"
      8000,       200,         2,    14.0,pymatching,deadbeef3,"{""d"":3}"
     10000,       200,         0,    22.0,pymatching,deadbeef4,"{""d"":5}"
"""

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                "--order",
                "metadata",
                str(d / "input.csv"),
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
      8000,       200,         2,    14.0,pymatching,deadbeef3,"{""d"":3}"
     10000,       200,         0,    22.0,pymatching,deadbeef4,"{""d"":5}"
      4000,       600,         6,    6.00,pymatching,deadbeef1,"{""d"":9}"
      2000,       200,         8,    4.00,pymatching,deadbeef0,"{""d"":19}"
      6000,       400,      4000,    10.0,pymatching,deadbeef2,"{""d"":200}"
"""

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                "--order",
                "error",
                str(d / "input.csv"),
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
     10000,       200,         0,    22.0,pymatching,deadbeef4,"{""d"":5}"
      8000,       200,         2,    14.0,pymatching,deadbeef3,"{""d"":3}"
      2000,       200,         8,    4.00,pymatching,deadbeef0,"{""d"":19}"
      4000,       600,         6,    6.00,pymatching,deadbeef1,"{""d"":9}"
      6000,       400,      4000,    10.0,pymatching,deadbeef2,"{""d"":200}"
"""



import dataclasses

from sinter._anon_task_stats import AnonTaskStats
from sinter._json_type import JSON_TYPE
from sinter._csv_out import csv_line


@dataclasses.dataclass(frozen=True)
class TaskStats:
    """Results of sampling from a decoding problem."""

    # Information describing the problem that was sampled.
    strong_id: str
    decoder: str
    json_metadata: JSON_TYPE

    # Information describing the results of sampling.
    shots: int
    errors: int
    discards: int
    seconds: float

    def __add__(self, other: 'TaskStats') -> 'TaskStats':
        assert self.strong_id == other.strong_id
        return TaskStats(
            decoder=self.decoder,
            strong_id=self.strong_id,
            json_metadata=self.json_metadata,
            shots=self.shots + other.shots,
            errors=self.errors + other.errors,
            discards=self.discards + other.discards,
            seconds=self.seconds + other.seconds,
        )

    def to_anon_stats(self) -> AnonTaskStats:
        return AnonTaskStats(
            shots=self.shots,
            errors=self.errors,
            discards=self.discards,
            seconds=self.seconds,
        )

    def to_csv_line(self) -> str:
        return csv_line(
            shots=self.shots,
            errors=self.errors,
            seconds=self.seconds,
            discards=self.discards,
            strong_id=self.strong_id,
            decoder=self.decoder,
            json_metadata=self.json_metadata)

    def __str__(self):
        return self.to_csv_line()



import math
import pathlib
from typing import Callable, List, TYPE_CHECKING
from typing import Tuple

import numpy as np
import stim

from sinter._decoding_decoder_class import Decoder

if TYPE_CHECKING:
    import fusion_blossom


class FusionBlossomDecoder(Decoder):
    """Use fusion blossom to predict observables from detection events."""
    def decode_via_files(self,
                         *,
                         num_shots: int,
                         num_dets: int,
                         num_obs: int,
                         dem_path: pathlib.Path,
                         dets_b8_in_path: pathlib.Path,
                         obs_predictions_b8_out_path: pathlib.Path,
                         tmp_dir: pathlib.Path,
                       ) -> None:
        try:
            import fusion_blossom
        except ImportError as ex:
            raise ImportError(
                "The decoder 'fusion_blossom' isn't installed\n"
                "To fix this, install the python package 'fusion-blossom' into your environment.\n"
                "For example, if you are using pip, run `pip install fusion-blossom~=0.1.4`.\n"
            ) from ex

        error_model = stim.DetectorErrorModel.from_file(dem_path)
        solver, fault_masks = detector_error_model_to_fusion_blossom_solver_and_fault_masks(error_model)
        num_det_bytes = math.ceil(num_dets / 8)

        with open(dets_b8_in_path, 'rb') as dets_in_f:
            with open(obs_predictions_b8_out_path, 'wb') as obs_out_f:
                for _ in range(num_shots):
                    dets_bit_packed = np.fromfile(dets_in_f, dtype=np.uint8, count=num_det_bytes)
                    if dets_bit_packed.shape != (num_det_bytes,):
                        raise IOError('Missing dets data.')
                    dets_sparse = np.flatnonzero(np.unpackbits(dets_bit_packed, count=num_dets, bitorder='little'))
                    syndrome = fusion_blossom.SyndromePattern(syndrome_vertices=dets_sparse)
                    solver.solve(syndrome)
                    prediction = int(np.bitwise_xor.reduce(fault_masks[solver.subgraph()]))
                    obs_out_f.write(prediction.to_bytes((num_obs + 7) // 8, byteorder='little'))
                    solver.clear()


def iter_flatten_model(model: stim.DetectorErrorModel,
                       handle_error: Callable[[float, List[int], List[int]], None],
                       handle_detector_coords: Callable[[int, np.ndarray], None]):
    det_offset = 0
    coords_offset = np.zeros(100, dtype=np.float64)

    def _helper(m: stim.DetectorErrorModel, reps: int):
        nonlocal det_offset
        nonlocal coords_offset
        for _ in range(reps):
            for instruction in m:
                if isinstance(instruction, stim.DemRepeatBlock):
                    _helper(instruction.body_copy(), instruction.repeat_count)
                elif isinstance(instruction, stim.DemInstruction):
                    if instruction.type == "error":
                        dets: List[int] = []
                        frames: List[int] = []
                        t: stim.DemTarget
                        p = instruction.args_copy()[0]
                        for t in instruction.targets_copy():
                            if t.is_relative_detector_id():
                                dets.append(t.val + det_offset)
                            elif t.is_logical_observable_id():
                                frames.append(t.val)
                            elif t.is_separator():
                                # Treat each component of a decomposed error as an independent error.
                                # (Ideally we could configure some sort of correlated analysis; oh well.)
                                handle_error(p, dets, frames)
                                frames = []
                                dets = []
                        # Handle last component.
                        handle_error(p, dets, frames)
                    elif instruction.type == "shift_detectors":
                        det_offset += instruction.targets_copy()[0]
                        a = np.array(instruction.args_copy())
                        coords_offset[:len(a)] += a
                    elif instruction.type == "detector":
                        a = np.array(instruction.args_copy())
                        for t in instruction.targets_copy():
                            handle_detector_coords(t.val + det_offset, a + coords_offset[:len(a)])
                    elif instruction.type == "logical_observable":
                        pass
                    else:
                        raise NotImplementedError()
                else:
                    raise NotImplementedError()
    _helper(model, 1)


def detector_error_model_to_fusion_blossom_solver_and_fault_masks(model: stim.DetectorErrorModel) -> Tuple['fusion_blossom.SolverSerial', np.ndarray]:
    """Convert a stim error model into a NetworkX graph."""

    import fusion_blossom

    def handle_error(p: float, dets: List[int], frame_changes: List[int]):
        if p == 0:
            return
        if len(dets) == 0:
            # No symptoms for this error.
            # Code probably has distance 1.
            # Accept it and keep going, though of course decoding will probably perform terribly.
            return
        if len(dets) == 1:
            dets = [dets[0], num_detectors]
        if len(dets) > 2:
            raise NotImplementedError(
                f"Error with more than 2 symptoms can't become an edge or boundary edge: {dets!r}.")
        if p > 0.5:
            # fusion_blossom doesn't support negative edge weights.
            # approximate them as weight 0.
            p = 0.5
        weight = math.log((1 - p) / p)
        mask = sum(1 << k for k in frame_changes)
        edges.append((dets[0], dets[1], weight, mask))

    def handle_detector_coords(detector: int, coords: np.ndarray):
        pass

    num_detectors = model.num_detectors
    edges: List[Tuple[int, int, float, int]] = []
    iter_flatten_model(
        model,
        handle_error=handle_error,
        handle_detector_coords=handle_detector_coords,
    )
    max_weight = max(1e-4, max((w for _, _, w, _ in edges), default=1))
    rescaled_edges = [
        (a, b, round(w * 2**10 / max_weight) * 2)
        for a, b, w, _ in edges
    ]
    fault_masks = np.array([e[3] for e in edges], dtype=np.uint64)

    initializer = fusion_blossom.SolverInitializer(
        num_detectors + 1,  # Total number of nodes.
        rescaled_edges,  # Weighted edges.
        [num_detectors],  # Boundary node.
    )

    return fusion_blossom.SolverSerial(initializer), fault_masks



import argparse
import json

import sys
from typing import List, Any

import sinter
from sinter._csv_out import CSV_HEADER
from sinter._existing_data import ExistingData
from sinter._plotting import better_sorted_str_terms


def main_combine(*, command_line_args: List[str]):
    parser = argparse.ArgumentParser()
    parser.add_argument('--order',
                        choices=('preserve', 'metadata', 'error'),
                        default='metadata',
                        help='Determines the order of output rows.\n'
                             '    metadata (default): sort ascending by metadata.'
                             '    preserve: match order of input rows.\n'
                             '    error: sort ascending by error rate')
    parser.add_argument('rest',
                        nargs=argparse.REMAINDER,
                        type=str,
                        help='Paths to CSV files containing sample statistics.')
    args = parser.parse_args(command_line_args)

    if args.rest:
        total = ExistingData()
        for path in args.rest:
            total += ExistingData.from_file(path)
    else:
        total = ExistingData.from_file(sys.stdin)

    if args.order == 'metadata':
        output = sorted(total.data.values(), key=lambda e: better_sorted_str_terms(json.dumps(e.json_metadata, separators=(',', ':'), sort_keys=True)))
    elif args.order == 'preserve':
        output = list(total.data.values())
    elif args.order == 'error':
        def err_rate_key(stats: sinter.TaskStats) -> Any:
            num_kept = stats.shots - stats.discards
            err_rate = 0 if num_kept == 0 else stats.errors / num_kept
            discard_rate = 0 if stats.shots == 0 else stats.discards / stats.shots
            return err_rate, discard_rate
        output = sorted(total.data.values(), key=err_rate_key)
    else:
        raise NotImplementedError(f'order={args.order}')

    print(CSV_HEADER)
    for value in output:
        print(value.to_csv_line())



import pathlib
import tempfile

import stim

import pytest
import sinter
from sinter._main import main
from sinter._main_combine import ExistingData
from sinter._plotting import split_by


def test_split_by():
    assert split_by('abcdefcccghi', lambda e: e == 'c') == [list('ab'), list('c'), list('def'), list('ccc'), list('ghi')]


def test_main_collect():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        for distance in [3, 5, 7]:
            c = stim.Circuit.generated(
                'repetition_code:memory',
                rounds=3,
                distance=distance,
                after_clifford_depolarization=0.02)
            with open(d / f'{distance}.stim', 'w') as f:
                print(c, file=f)

        # Collects requested stats.
        main(command_line_args=[
            "collect",
            "--circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "--max_shots",
            "1000",
            "--max_errors",
            "10",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        data = ExistingData.from_file(d / "out.csv").data
        assert len(data) == 3
        for k, v in data.items():
            assert v.discards == 0
            assert v.errors <= 50
            assert v.shots >= 1000

        # No more work when existing stats at merge location are sufficient.
        with open(d / "out.csv") as f:
            contents1 = f.read()
        main(command_line_args=[
            "collect",
            "--circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "--max_shots",
            "1000",
            "--max_errors",
            "10",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        with open(d / "out.csv") as f:
            contents2 = f.read()
        assert contents1 == contents2

        # No more work when existing work is sufficient.
        main(command_line_args=[
            "collect",
            "--circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "--max_shots",
            "1000",
            "--max_errors",
            "10",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--existing_data_filepaths",
            str(d / "out.csv"),
            "--save_resume_filepath",
            str(d / "out2.csv"),
        ])
        data2 = ExistingData.from_file(d / "out2.csv").data
        assert len(data2) == 0


class AlternatingPredictionsDecoder(sinter.Decoder):
    def decode_via_files(self,
                         *,
                         num_shots: int,
                         num_dets: int,
                         num_obs: int,
                         dem_path: pathlib.Path,
                         dets_b8_in_path: pathlib.Path,
                         obs_predictions_b8_out_path: pathlib.Path,
                         tmp_dir: pathlib.Path,
                       ) -> None:
        bytes_per_shot = (num_obs + 7) // 8
        with open(obs_predictions_b8_out_path, 'wb') as f:
            for k in range(num_shots):
                f.write((k % 3 == 0).to_bytes(length=bytes_per_shot, byteorder='little'))


def _make_custom_decoders():
    return {'alternate': AlternatingPredictionsDecoder()}


def test_main_collect_with_custom_decoder():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'tmp.stim', 'w') as f:
            print("""
                M(0.1) 0
                DETECTOR rec[-1]
                OBSERVABLE_INCLUDE(0) rec[-1]
            """, file=f)

        with pytest.raises(ValueError, match="Not a recognized decoder"):
            main(command_line_args=[
                "collect",
                "--circuits",
                str(d / "tmp.stim"),
                "--max_shots",
                "1000",
                "--decoders",
                "NOTEXIST",
                "--custom_decoders_module_function",
                "sinter._main_test:_make_custom_decoders",
                "--processes",
                "2",
                "--quiet",
                "--save_resume_filepath",
                str(d / "out.csv"),
            ])

        # Collects requested stats.
        main(command_line_args=[
            "collect",
            "--circuits",
            str(d / "tmp.stim"),
            "--max_shots",
            "1000",
            "--decoders",
            "alternate",
            "--custom_decoders_module_function",
            "sinter._main_test:_make_custom_decoders",
            "--processes",
            "2",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        data = ExistingData.from_file(d / "out.csv").data
        assert len(data) == 1
        v, = data.values()
        assert v.shots == 1000
        assert 50 < v.errors < 500
        assert v.discards == 0


def test_main_collect_post_select_observables():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'circuit.stim', 'w') as f:
            print("""
                M(0.125) 0 1
                OBSERVABLE_INCLUDE(0) rec[-1]
                OBSERVABLE_INCLUDE(11) rec[-1] rec[-2]
            """, file=f)

        # Collects requested stats.
        main(command_line_args=[
            "collect",
            "--postselected_observables_predicate",
            "index == 11",
            "--circuits",
            str(d / "circuit.stim"),
            "--max_shots",
            "10000",
            "--max_errors",
            "10000",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        data = sinter.stats_from_csv_files(d / "out.csv")
        assert len(data) == 1
        stats, = data
        assert stats.shots == 10000
        assert 0.21875 - 0.1 < stats.discards / stats.shots < 0.21875 + 0.1
        assert 0.015625 - 0.01 <= stats.errors / (stats.shots - stats.discards) <= 0.015625 + 0.02


def test_main_collect_comma_separated_key_values():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        paths = []
        for distance in [3, 5, 7]:
            c = stim.Circuit.generated(
                'repetition_code:memory',
                rounds=3,
                distance=distance,
                after_clifford_depolarization=0.02)
            path = d / f'd={distance},p=0.02,r=3.0,c=rep_code.stim'
            paths.append(str(path))
            with open(path, 'w') as f:
                print(c, file=f)

        # Collects requested stats.
        main(command_line_args=[
            "collect",
            "--circuits",
            *paths,
            "--max_shots",
            "1000",
            "--metadata_func",
            "sinter.comma_separated_key_values(path)",
            "--max_errors",
            "10",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        data = sinter.stats_from_csv_files(d / "out.csv")
        seen_metadata = frozenset(repr(e.json_metadata) for e in data)
        assert seen_metadata == frozenset([
            "{'c': 'rep_code', 'd': 3, 'p': 0.02, 'r': 3.0}",
            "{'c': 'rep_code', 'd': 5, 'p': 0.02, 'r': 3.0}",
            "{'c': 'rep_code', 'd': 7, 'p': 0.02, 'r': 3.0}",
        ])


def test_main_predict():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.dets', 'w') as f:
            print("""
shot D0
shot
            """, file=f)
        with open(d / f'input.dem', 'w') as f:
            print("""
error(0.1) D0 L0
            """, file=f)

        main(command_line_args=[
            "predict",
            "--dets",
            str(d / "input.dets"),
            "--dem",
            str(d / "input.dem"),
            "--decoder",
            "pymatching",
            "--dets_format",
            "dets",
            "--obs_out",
            str(d / "output.01"),
            "--obs_out_format",
            "01",
        ])
        with open(d / 'output.01') as f:
            assert f.read() == '1\n0\n'



import pathlib
import tempfile

import numpy as np
import pytest
import stim

from sinter._collection import post_selection_mask_from_4th_coord
from sinter._decoding import sample_decode
from sinter._decoding_all_built_in_decoders import BUILT_IN_DECODERS

DECODER_PACKAGES = [
    ('pymatching', 'pymatching'),
    ('fusion_blossom', 'fusion_blossom'),
    ('internal', 'gqec'),
    ('internal_correlated', 'gqec'),
    ('internal_2', 'gqec'),
    ('internal_correlated_2', 'gqec'),
]


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_decode_repetition_code(decoder: str, required_import: str):
    pytest.importorskip(required_import)

    circuit = stim.Circuit.generated('repetition_code:memory',
                                     rounds=3,
                                     distance=3,
                                     after_clifford_depolarization=0.05)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert 1 <= result.errors <= 100
    assert result.shots == 1000


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_decode_surface_code(decoder: str, required_import: str):
    pytest.importorskip(required_import)

    circuit = stim.Circuit.generated(
        "surface_code:rotated_memory_x",
        distance=3,
        rounds=15,
        after_clifford_depolarization=0.001,
    )
    stats = sample_decode(
        num_shots=1000,
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        decoder=decoder,
    )
    assert 0 <= stats.errors <= 50


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_empty(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit()
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert result.errors == 0


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_no_observables(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        DETECTOR rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert result.errors == 0


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_invincible_observables(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0 1
        DETECTOR rec[-2]
        OBSERVABLE_INCLUDE(1) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert result.errors == 0


@pytest.mark.parametrize('decoder,required_import,offset', [(a, b, c) for a, b in DECODER_PACKAGES for c in range(8)])
def test_observable_offsets_mod8(decoder: str, required_import: str, offset: int):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        MR 0
        DETECTOR rec[-1]
    """) * (8 + offset) + stim.Circuit("""
        X_ERROR(0.1) 0
        MR 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert 50 <= result.errors <= 150


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_no_detectors(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert 50 <= result.errors <= 150


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_no_detectors_with_post_mask(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        post_mask=np.array([], dtype=np.uint8),
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert 50 <= result.errors <= 150


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_post_selection(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.6) 0
        M 0
        DETECTOR(2, 0, 0, 1) rec[-1]
        OBSERVABLE_INCLUDE(0) rec[-1]

        X_ERROR(0.5) 1
        M 1
        DETECTOR(1, 0, 0) rec[-1]
        OBSERVABLE_INCLUDE(0) rec[-1]
        
        X_ERROR(0.1) 2
        M 2
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        post_mask=post_selection_mask_from_4th_coord(circuit),
        num_shots=2000,
        decoder=decoder,
    )
    assert 1050 <= result.discards <= 1350
    assert 40 <= result.errors <= 160


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_decode_fails_correctly(decoder: str, required_import: str):
    pytest.importorskip(required_import)

    decoder_obj = BUILT_IN_DECODERS.get(decoder)
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        circuit = stim.Circuit("""
            REPEAT 9 {
                MR(0.001) 0
                DETECTOR rec[-1]
                OBSERVABLE_INCLUDE(0) rec[-1]
            }
        """)
        dem = circuit.detector_error_model()
        circuit.to_file(d / 'circuit.stim')
        dem.to_file(d / 'dem.dem')
        with open(d / 'bad_dets.b8', 'wb') as f:
            f.write(b'!')

        with pytest.raises(Exception):
            decoder_obj.decode_via_files(
                num_shots=1,
                num_dets=dem.num_detectors,
                num_obs=dem.num_observables,
                dem_path=d / 'dem.dem',
                dets_b8_in_path=d / 'bad_dets.b8',
                obs_predictions_b8_out_path=d / 'predict.b8',
                tmp_dir=d,
            )



import abc

import pathlib


class Decoder(metaclass=abc.ABCMeta):
    """Predicts observable flips from detection event data.

    Custom decoders can be explained to sinter by inheriting from this class and
    implementing its methods in a way that call the custom decoder.

    Decoder classes MUST be serializable (e.g. via pickling), so that they can be given to
    worker processes when using python multiprocessing.
    """

    @abc.abstractmethod
    def decode_via_files(self,
                         *,
                         num_shots: int,
                         num_dets: int,
                         num_obs: int,
                         dem_path: pathlib.Path,
                         dets_b8_in_path: pathlib.Path,
                         obs_predictions_b8_out_path: pathlib.Path,
                         tmp_dir: pathlib.Path,
                       ) -> None:
        """Performs decoding by reading problems from, and writing solutions to, file paths.

        Args:
            num_shots: The number of times the circuit was sampled. The number of problems
                to be solved.
            num_dets: The number of detectors in the circuit. The number of detection event
                bits in each shot.
            num_obs: The number of observables in the circuit. The number of predicted bits
                in each shot.
            dem_path: The file path where the detector error model should be read from,
                e.g. using `stim.DetectorErrorModel.from_file`. The error mechanisms
                specified by the detector error model should be used to configure the
                decoder.
            dets_b8_in_path: The file path that detection event data should be read from.
                Note that the file may be a named pipe instead of a fixed size object.
                The detection events will be in b8 format (see
                https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md ). The
                number of detection events per shot is available via the `num_dets`
                argument or via the detector error model at `dem_path`.
            obs_predictions_b8_out_path: The file path that decoder predictions must be
                written to. The predictions must be written in b8 format (see
                https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md ). The
                number of observables per shot is available via the `num_obs` argument or
                via the detector error model at `dem_path`.
            tmp_dir: Any temporary files generated by the decoder during its operation MUST
                be put into this directory. The reason for this requirement is because
                sinter is allowed to kill the decoding process without warning, without
                giving it time to clean up any temporary objects. All cleanup should be done
                via sinter deleting this directory after killing the decoder.
        """
        pass



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

with open('README.md', encoding='UTF-8') as f:
    long_description = f.read()
with open('requirements.txt', encoding='UTF-8') as f:
    requirements = [line.split()[0] for line in f.read().splitlines()]

version = '1.11.dev1675231726'

setup(
    name='sinter',
    version=version,
    author='Craig Gidney',
    author_email='craig.gidney@gmail.com',
    license='Apache 2',
    packages=['sinter'],
    package_dir={'': 'src'},
    description='Samples stim circuits and decodes them using pymatching.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.7.0',
    data_files=['README.md', 'requirements.txt', 'readme_example_plot.png'],
    install_requires=requirements,
    tests_require=['pytest'],
    entry_points={
        'console_scripts': ['sinter=sinter._main:main'],
    },
)



import dataclasses


@dataclasses.dataclass(frozen=True)
class AnonTaskStats:
    """Results of sampling from a decoding problem."""

    shots: int = 0
    errors: int = 0
    discards: int = 0
    seconds: float = 0

    def __post_init__(self):
        assert isinstance(self.errors, int)
        assert isinstance(self.shots, int)
        assert isinstance(self.discards, int)
        assert isinstance(self.seconds, (int, float))
        assert self.errors >= 0
        assert self.discards >= 0
        assert self.seconds >= 0
        assert self.shots >= self.errors + self.discards

    def __add__(self, other: 'AnonTaskStats') -> 'AnonTaskStats':
        if not isinstance(other, AnonTaskStats):
            return NotImplemented
        return AnonTaskStats(
            shots=self.shots + other.shots,
            errors=self.errors + other.errors,
            discards=self.discards + other.discards,
            seconds=self.seconds + other.seconds,
        )
